<script src="{{ Lona.server.settings.STATIC_URL_PREFIX }}lona/lona.js"></script>
{% for script in scripts %}<script src="{{ Lona.server.settings.STATIC_URL_PREFIX }}{{ script.url }}"></script>
{% endfor -%}
